#!/usr/bin/env python
#
# bezier_loop.py v1.1
# Copyright 2008 Sean McKean <smckean AT yvn DOT com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import pygame
from pygame.locals import *
from random import random, randrange
from math import cos, pi

# Globals, described in the usage string near the end of the program.
ScreenSize = 256
ScreenFlags = 0
NumberBeziers = 4
NumberDivisions = 32
TargetFPS = 60
MinAlpha = 0
MaxAlpha = 255
AngleDelta = 0.01
# Common colors
white = 255, 255, 255
black = 0, 0, 0
FadeOutColor = white

def CreateSequenceSfc(size, color_name_a, color_name_b):
    """Creates and returns an n*n surface depicting a 2-dimensional
       sequence.
    """
    power = 1
    while 2 ** power <= size:
        power += 1
    power -= 1
    size = 2 ** power
    sequence = [0]
    for p in range(power * 2):
        sequence = sequence + [1 - q for q in sequence]
    color_a = ''.join([chr(pygame.Color(color_name_a)[q]) for q in range(3)])
    color_b = ''.join([chr(pygame.Color(color_name_b)[q]) for q in range(3)])
    image_str = ''
    for q in range(size ** 2):
        if sequence[q] == 1:
            image_str += color_a
        else:
            image_str += color_b
    sfc = pygame.image.frombuffer(image_str, (size, size), 'RGB')
    return sfc

class BezierLoop:
    """Creates and updates/draws a loop consisting of connected bezier
       curves.
    """

    def __init__(self, n_points, n_subdivisions, sfc):
        """Each curve consists of 2 endpoints and 2 control points.
           To make the curves appear to move, the endpoints and control
           points are adjusted every frame by a corresponding delta
           value.
        """
        self.divs = n_subdivisions
        self.sfc = sfc
        self.s_width, self.s_height = self.sfc.get_size()
        self.end_points = []
        self.end_deltas = []
        self.con_points = []
        self.con_deltas = []
        for q in range(n_points):
            position = [randrange(self.s_width), randrange(self.s_height)]
            deltas = [ (random() / 2 + 0.5) * (random() < 0.5) and -1 or 1,
                       (random() / 2 + 0.5) * (random() < 0.5) and -1 or 1 ]
            self.end_points.append(position)
            self.end_deltas.append(deltas)
            position = [randrange(self.s_width), randrange(self.s_height)]
            deltas = [ (random() / 2 + 0.5) * (random() < 0.5) and -1 or 1,
                       (random() / 2 + 0.5) * (random() < 0.5) and -1 or 1 ]
            self.con_points.append(position)
            self.con_deltas.append(deltas)

    def CalculateSegments(self):
        """Creates and returns the bezier loop points by subdividing
           each curve and extending the points list. Check the Wikipedia
           article on bezier curves for more information.
        """
        shift_ends = self.end_points[1: ] + self.end_points[: 1]
        shift_cons = self.con_points[1: ] + self.con_points[: 1]
        points = []
        for e1, c1, c2, e2 in zip( self.end_points, self.con_points,
                                   shift_cons, shift_ends ):
            c2 = [2 * e2[0] - c2[0], 2 * e2[1] - c2[1]]
            bezier_pts = []
            for q in range(self.divs):
                minus = float(self.divs - q) / self.divs
                ratio = float(q) / self.divs
                bezier_pts.append(( minus ** 3 * e1[0] +
                                     3 * minus ** 2 * ratio * c1[0] +
                                     3 * minus * ratio ** 2 * c2[0] +
                                     ratio ** 3 * e2[0],
                                    minus ** 3 * e1[1] +
                                     3 * minus ** 2 * ratio * c1[1] +
                                     3 * minus * ratio ** 2 * c2[1] +
                                     ratio ** 3 * e2[1]))
            points.extend(bezier_pts)
        return points

    def Draw(self):
        """Draws the bezier loop by calling draw.lines function,
           which speeds things up considerably.
           x_off and y_off give the loop different offsets to draw with
           so that each drawn frame appears continuous with the
           previous one.
        """
        self.sfc.fill(black)
        points = self.CalculateSegments()
        for x_off, y_off in ( (0, 0), (1, 0), (2, 0), (-1, 0), (-2, 0),
                              (0, 1), (0, 2), (0, -1), (0, -2) ):
            points_off = [(x + x_off, y + y_off) for x, y in points]
            pygame.draw.lines(self.sfc, white, True, points_off)

    def UpdatePoint(self, point, delta):
        """Moves the given point across the screen by the given delta,
           and negates the delta if a border is crossed.
        """
        point[0] += delta[0]
        if point[0] < 0:
            point[0] = 0.0
            delta[0] *= -1
        elif point[0] >= self.s_width:
            point[0] = self.s_width - 1.0
            delta[0] *= -1
        point[1] += delta[1]
        if point[1] < 0:
            point[1] = 0.0
            delta[1] *= -1
        elif point[1] >= self.s_height:
            point[1] = self.s_height - 1.0
            delta[1] *= -1

    def Update(self):
        """Calls self.UpdatePoint for each corresponding endpoint/delta
           or control point/delta.
        """
        for epoint, edelta in zip(self.end_points, self.end_deltas):
            self.UpdatePoint(epoint, edelta)
        for cpoint, cdelta in zip(self.con_points, self.con_deltas):
            self.UpdatePoint(cpoint, cdelta)

class Main:
    """Handles initialization code and frame drawing/control."""

    def __init__(self, fname, size, flags):
        if fname == '' or not os.path.exists(fname):
            self.back_sfc = CreateSequenceSfc(size, 'cyan', 'black')
        else:
            self.back_sfc = pygame.image.load(fname)
        size = self.back_sfc.get_size()
        self.s_width, self.s_height = size
        pygame.init()
        self.screen = pygame.display.set_mode(size, flags, 32)
        self.clock = pygame.time.Clock()
        self.back_sfc = self.back_sfc.convert()
        self.final_sfc = pygame.Surface(size).convert()
        self.final_sfc.set_colorkey(black)
        self.bez_sfc = pygame.Surface(size).convert()
        self.bez_sfc.set_colorkey(white)
        self.beziers_list = BezierLoop( NumberBeziers, NumberDivisions,
                                        self.bez_sfc )
        self.blur_sfc = pygame.Surface(size).convert()
        self.blur_sfc.fill(black)
        # Set alpha channel to a nice blending value.
        self.blur_sfc.set_alpha(4)
        self.back_alpha_sfc = pygame.Surface(size).convert()
        self.back_alpha_sfc.fill(FadeOutColor)
        self.back_blit_sfc = pygame.Surface(size).convert()
        self.midpoint = float(MaxAlpha + MinAlpha) / 2 + 0.1
        self.amplitude = float(MaxAlpha - MinAlpha) / 2
        self.angle = 0.0
        self.UpdateAlphaValue()

    def GetBackgroundAlpha(self):
        """Returns the alpha value according to function cosine."""
        return int(self.midpoint + self.amplitude * cos(self.angle))

    def UpdateAlphaValue(self):
        """Alpha value moves according to function
           self.GetBackgroundAlpha(). Angle is increased according to
           AngleDelta.
        """
        self.back_alpha = self.GetBackgroundAlpha()
        self.angle = (self.angle + AngleDelta) % (2 * pi)

    def DrawMyriadSfcs(self):
        """Yes, this is a lot of full-frame blits, but it is necessary
           to achieve the blending of the back surface with the bezier
           loop, and the blurring of each frame.
        """
        self.screen.blit(self.blur_sfc, (0, 0))
        self.back_blit_sfc.blit(self.back_sfc, (0, 0))
        self.back_alpha_sfc.set_alpha(int(self.back_alpha))
        self.back_blit_sfc.blit(self.back_alpha_sfc, (0, 0))
        self.final_sfc.blit(self.back_blit_sfc, (0, 0))
        self.beziers_list.Draw()
        self.final_sfc.blit(self.bez_sfc, (0, 0))
        self.screen.blit(self.final_sfc, (0, 0))

    def MainLoop(self):
        """Handles input (currently exits on keyboard/quit event),
           draws surfaces, flips framebuffer, and updates values for
           next frame.
        """
        while True:
            for event in pygame.event.get():
                if event.type in (pygame.QUIT, pygame.KEYDOWN):
                    return
            self.DrawMyriadSfcs()
            pygame.display.flip()
            self.clock.tick(TargetFPS)
            self.beziers_list.Update()
            self.UpdateAlphaValue()

if __name__ == '__main__':

    Usage = """\nUsage: python %s [filename] [-global=value] [-h]

-h : Displays usage information and exits.
filename : an image that will be displayed in the background.
If omitted, a default surface will be provided.
-global : Sets a variable name in the following list to the given value.
Valid global variables:

-ScreenSize      -- Sets screen size (n*n) when filename is not provided.
                    =%d
-ScreenFlags     -- Try pygame constants, such as FULLSCREEN
                    =%s
-NumberBeziers   -- Sets number of curves in the drawn loop.
                    =%s
-NumberDivisions -- Sets how many divisions are in each bezier curve
                    (set to a higher value for more curviness).
                    =%s
-TargetFPS       -- Program will try not to surpass given FPS.
                    If set to 0, no time-delaying will be performed.
                    =%s
-MinAlpha        -- The alpha value (0-255) which indicates the highest
                    translucency of the alpha surface.
                    =%s
-MaxAlpha        -- The alpha value (0-255) which indicates the highest
                    opaqueness of the alpha surface.
                    =%s
-AngleDelta      -- How quickly the alpha value cycles.
                    =%s
-FadeOutColor    -- The RGB value that the alpha surface will display.
                    =%s
"""
    if hasattr(FadeOutColor, '__len__'):
        fade_out_string = str(FadeOutColor).replace(' ', '')
    elif type(FadeOutColor) == int:
        fade_out_string = hex(FadeOutColor)
    else:
        fade_out_string = str(FadeOutColor)
    Usage %= ( sys.argv[0], ScreenSize, ScreenFlags, NumberBeziers,
               NumberDivisions, TargetFPS, MinAlpha, MaxAlpha, AngleDelta,
               fade_out_string )
    fname = ''
    for arg in sys.argv[1: ]:
        if arg.startswith('-'):
            rest = arg[1: ]
            if rest.startswith('h'):
                print Usage
                sys.exit(0)
            else:
                for global_var in ( 'ScreenSize', 'ScreenFlags',
                                    'NumberBeziers', 'NumberDivisions',
                                    'TargetFPS', 'MinAlpha', 'MaxAlpha',
                                    'AngleDelta', 'FadeOutColor'):
                    if rest.lower().startswith(global_var.lower()):
                        try:
                            # rest[].upper() is for automatically changing
                            # given pygame constants to uppercase.
                            exec ''.join( ( global_var,
                                            rest[rest.find('='): ].upper() ) )
                        except:
                            print 'Invalid global value given!'
                            print 'Give -h option for list of options.'
                            sys.exit(1)
        elif os.path.exists(arg):
            fname = arg
        else:
            print Usage
            sys.exit(1)
    main = Main(fname, ScreenSize, ScreenFlags)
    main.MainLoop()
    print "Average FPS: %.2f" % main.clock.get_fps()
